from typing import List
import os
import sys
import json
import subprocess
from threading import Thread
import signal
from pathlib import Path
import socket
import importlib
from .handle_request import handle_request
from .RtcshareContext import RtcshareContext

this_directory = Path(__file__).parent

class Daemon:
    def __init__(self):
        self.process = None
        self.output_thread = None
        self.context = RtcshareContext()
    
    def initialize_plugins(self, plugin_names: List[str]):
        for plugin_name in plugin_names:
            module = importlib.import_module(plugin_name)
            plugin = getattr(module, "RtcsharePlugin")
            plugin.initialize(self.context)

    def _forward_output(self):
        while True:
            line = self.process.stdout.readline()
            sys.stdout.write(line)
            sys.stdout.flush()
            return_code = self.process.poll()
            if return_code is not None:
                print(f'Process exited with return code {return_code}')
                break
    
    def _handle_client(self, client_socket: socket.socket):
        received_data = b''
        while True:
            data0 = client_socket.recv(4096)
            received_data += data0
            if data0.endswith(b'\n'):
                break
        request = json.loads(received_data.decode('utf-8'))

        # Process data received from rtcshare-js and send a response
        try:
            response_data = handle_request(request, self.context)
        except Exception as e:
            error_string = str(e)
            response_data = None
        if response_data is not None:
            client_socket.sendall(response_data)
            client_socket.close()
        else:
            client_socket.sendall(b'\n' + error_string.encode('utf-8'))
            client_socket.close()

    def start_socket_server(self) -> int:
        self.socket_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket_server.bind(("localhost", 0))
        self.socket_server.listen(1)

        self.socket_server_thread = Thread(target=self._accept_connections, daemon=True) # daemon=True means that the thread will not block the program from exiting
        self.socket_server_thread.start()

        return self.socket_server.getsockname()[1]  # Return the port number
    
    def stop_socket_server(self):
        self.socket_server.close()
    
    def _accept_connections(self):
        while True:
            client_socket, client_address = self.socket_server.accept()
            client_thread = Thread(target=self._handle_client, args=(client_socket,), daemon=True) # daemon=True means that the thread will not block the program from exiting
            client_thread.start()

    def _handle_exit(self, signum, frame):
        print('Exiting')
        self.stop()
        sys.exit(0)

    def start(self, *, enable_remote_access: bool = False):
        socket_server_port = self.start_socket_server()
        os.environ["RTCSHARE_SOCKET_PORT"] = str(socket_server_port)  # Pass the port number to the js server

        dir0 = os.environ.get('RTCSHARE_DIR')

        # apparently shell=True is necessary for Windows, but shell=False is necessary for Linux
        if os.name == 'nt':
            shell = True
        elif os.name == 'posix':
            shell = False
        else:
            print(f'Warning: unrecognized os.name: {os.name}')
            shell = False

        try:
            npm_version = subprocess.run(["npm", "--version"], stdout=subprocess.PIPE, universal_newlines=True, shell=shell).stdout.strip()
            print(f'npm version: {npm_version}')
        except:
            raise Exception('Unable to run npm.')
        
        try:
            node_version = subprocess.run(["node", "--version"], stdout=subprocess.PIPE, universal_newlines=True, shell=shell).stdout.strip()
            print(f'node version: {node_version}')
        except:
            raise Exception('Unable to run node.')
        
        # parse node_version v18.0.0 to get the major version number
        node_major_version = int(node_version.split('.')[0][1:])
        if node_major_version < 16:
            raise Exception('node version must be >= 16.0.0')

        # run the command npm install in the js directory
        subprocess.run(["npm", "install"], cwd=f'{this_directory}/js', shell=shell)

        # run the build command
        subprocess.run(["npm", "run", "build"], cwd=f'{this_directory}/js', shell=shell)

        cmd = ["node", f'{this_directory}/js/dist/index.js', "start", "--dir", dir0, "--verbose"]
        if enable_remote_access:
            cmd.append("--enable-remote-access")
        self.process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
            universal_newlines=True,
        )

        self.output_thread = Thread(target=self._forward_output, daemon=True) # daemon=True means that the thread will not block the program from exiting
        self.output_thread.start()

        signal.signal(signal.SIGINT, self._handle_exit)
        signal.signal(signal.SIGTERM, self._handle_exit)

    def stop(self):
        self.stop_socket_server()
        if self.process:
            self.process.terminate()
            self.process.wait()

def start(dir: str, *, enable_remote_access: bool = False, plugin_names: List[str] = []):
    os.environ['RTCSHARE_DIR'] = dir
    daemon = Daemon()
    daemon.initialize_plugins(plugin_names)
    daemon.start(enable_remote_access=enable_remote_access)

    # Don't exit until the output thread exits
    daemon.output_thread.join()